import functools
import json
import importlib.resources as importlib_resources

__all__ = ["get_primary_meta_recipe",
           "get_secondary_meta_recipe",
           ]


@functools.lru_cache()
def get_primary_meta_recipe():
    ref = importlib_resources.files(
        "afmformats.formats.fmt_jpk") / "jpk_meta_primary.json"
    with importlib_resources.as_file(ref) as path:
        with open(path) as fd:
            meta_recipe_pri = json.load(fd)
    return meta_recipe_pri


@functools.lru_cache()
def get_secondary_meta_recipe():
    ref = importlib_resources.files(
        "afmformats.formats.fmt_jpk") / "jpk_meta_secondary.json"
    with importlib_resources.as_file(ref) as path:
        with open(path) as fd:
            meta_recipe_sec = json.load(fd)
    return meta_recipe_sec



"""Test .tab format functionalities"""
import pathlib
import tempfile

import numpy as np
import pytest

import afmformats


data_path = pathlib.Path(__file__).resolve().parent / "data"


def generate_tab_file():
    tf = tempfile.mktemp(suffix=".tab", prefix="afmformats_test_tab_")
    with open(tf, "w") as fd:
        fd.write('# BEGIN METADATA\r\n')
        fd.write('# {\r\n')
        fd.write('#   "curve id": "2016.06.15-10.18.28-00093:6",\r\n')
        fd.write('#   "date": "2016-06-15",\r\n')
        fd.write('#   "duration": 4.2,\r\n')
        fd.write('#   "enum": 0,\r\n')
        fd.write('#   "feedback mode": "contact",\r\n')
        fd.write('#   "format": "JPK Instruments (binary FD data)",\r\n')
        fd.write('#   "imaging mode": "force-distance",\r\n')
        fd.write('#   "instrument": "JPK00758-CellHesion-200",\r\n')
        fd.write('#   "point count": 8400,\r\n')
        fd.write('#   "position x": -0.0021247991071428572,\r\n')
        fd.write('#   "position y": -0.0025797712053571434,\r\n')
        fd.write('#   "rate approach": 2000.0,\r\n')
        fd.write('#   "rate retract": 2000.0,\r\n')
        fd.write('#   "sensitivity": 9.120708714532604e-08,\r\n')
        fd.write('#   "session id": "2016.06.15-10.18.28-00093",\r\n')
        fd.write('#   "setpoint": 6.000000000000001e-09,\r\n')
        fd.write('#   "software": "JPK",\r\n')
        fd.write('#   "software version": "5.0.76",\r\n')
        fd.write('#   "speed approach": 9.999999999999999e-06,\r\n')
        fd.write('#   "speed retract": 2.4999999999999998e-05,\r\n')
        fd.write('#   "spring constant": 0.02230065831827453,\r\n')
        fd.write('#   "time": "15:52:13.937",\r\n')
        fd.write('#   "z range": 2.6486991447509974e-05\r\n')
        fd.write('# }\r\n')
        fd.write('# END METADATA\r\n')
        fd.write('# force\ttip position\tsegment\r\n')
        for ii in range(100):
            fd.write("{}\t{}\t{}\r\n".format(ii*1e-9,
                                             np.pi*ii*1e-6,
                                             ii > 50))
    return tf


def test_detect():
    tf = generate_tab_file()
    recipe = afmformats.formats.get_recipe(tf)
    assert recipe.descr == "tab-separated values"


def test_detect_bad():
    _, tf = tempfile.mkstemp(suffix=".tab", prefix="afmformats_test")

    with pytest.raises(afmformats.errors.FileFormatNotSupportedError):
        afmformats.formats.get_recipe(tf)


def test_open_0_13_3():
    fdat = afmformats.load_data(data_path / "fmt-tab-fd_version_0.13.3.tab")[0]
    assert fdat.metadata["imaging mode"] == "force-distance"
    assert fdat["force"][2000] == -6.7762282e-10  # cropped compared to h5 file
    # set the tip position
    fdat["tip position"] = fdat["height (measured)"]*1.01


def test_open_simple():
    tf = generate_tab_file()
    data = afmformats.load_data(tf)[0]
    assert np.sum(data["segment"]) == 49
    assert np.allclose(data["force"][0], 0)
    assert np.allclose(data["force"][1], 1e-9)
    assert np.allclose(data["tip position"][1], np.pi*1e-6)


def test_save_open_with_metadata():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    fdat = afmformats.load_data(jpkfile, modality="force-distance")[0]
    _, path = tempfile.mkstemp(suffix=".tab", prefix="afmformats_test_")
    fdat.export_data(path, metadata=True, fmt="tab")
    fdat2 = afmformats.load_data(path, modality="force-distance")[0]
    # compare metadata
    for key in fdat.metadata:
        if key in ["path", "format"]:
            assert fdat.metadata[key] != fdat2.metadata[key]
        else:
            assert fdat.metadata[key] == fdat2.metadata[key]
    for col in fdat.columns:
        assert np.allclose(fdat[col], fdat2[col], atol=0)


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""Test group functionalities"""
import pathlib

import pytest

from afmformats import AFMForceDistance, AFMGroup, load_data


data_path = pathlib.Path(__file__).parent / "data"


def test_append_bad_type():
    grp = AFMGroup(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    with pytest.raises(ValueError, match="must be an instance"):
        grp.append(2)


def test_base():
    grp1 = AFMGroup(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    grp2 = load_data(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")

    grp3 = grp1 + grp2
    assert len(grp3) == 5
    assert len(grp2) == 1
    assert len(grp1) == 4

    grp2 += grp1
    assert len(grp3) == 5
    assert len(grp2) == 5

    for afmd in grp3:
        assert isinstance(afmd, AFMForceDistance)

    # test repr
    assert "AFMGroup" in repr(grp3)

    # test str
    assert "AFMGroup" in str(grp3)
    assert "fmt-jpk-fd_spot3-0192.jpk-force" in str(grp3)


def test_enum_index():
    grp = AFMGroup(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    afmd = grp.get_enum(2)
    assert afmd.enum == 2
    assert grp.index(afmd) == 2


def test_enum_multiple_error():
    grp = AFMGroup(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    grp += load_data(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")
    with pytest.raises(ValueError, match="Multiple curves"):
        grp.get_enum(0)


def test_enum_not_found_error():
    grp = AFMGroup(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    with pytest.raises(KeyError, match="Could not find"):
        grp.get_enum(100)


def test_iadd_bad_type():
    grp = AFMGroup(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    with pytest.raises(ValueError, match="AFMGroup for iadd"):
        grp += 2


def test_meta_override():
    k = 20
    sens = .01e-6
    grp = AFMGroup(
        data_path / "fmt-afm-workshop-fd_mapping_16_2018-08-01_13.07.zip",
        meta_override={"spring constant": k,
                       "sensitivity": sens})
    assert grp[0].metadata["spring constant"] == k


def test_meta_override_meaningless():
    with pytest.raises(ValueError, match="meaningless"):
        AFMGroup(meta_override={"spring constant": 1,
                                "sensitivity": 2})


def test_subgroup():
    group = AFMGroup()
    group += load_data(data_path /
                       "fmt-jpk-fd_map-data-reference-points.jpk-force-map")
    group += load_data(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    group += load_data(data_path /
                       "fmt-jpk-fd_flipsign_2015.05.22-15.31.49.352.jpk-force")

    exp = data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map"
    subgrp = group.subgroup_with_path(path=exp)
    assert len(group) == 8
    assert len(subgrp) == 4
    assert subgrp[0].path == exp



"""Test AFM workshop format"""
import pathlib
import tempfile

import numpy as np
import pytest

import afmformats


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_map_grid():
    tf = data_path / "fmt-afm-workshop-fd_mapping_16_2018-08-01_13.07.zip"
    k = 20
    sens = .01e-6
    qmap = afmformats.AFMQMap(tf, meta_override={"spring constant": k,
                                                 "sensitivity": sens})
    assert qmap.shape == (4, 4)
    data1 = qmap.group[1]
    assert data1.metadata["grid size x"] == 2.533333333333333e-05
    assert data1.metadata["grid size y"] == 2.533333333333333e-05
    assert data1.metadata["grid index x"] == 1
    assert data1.metadata["grid index y"] == 0
    assert data1.metadata["position x"] == 7.333333333333332e-06
    assert data1.metadata["position y"] == 2e-6
    assert data1.metadata["grid shape x"] == 4
    assert data1.metadata["grid shape y"] == 4
    assert data1.metadata["grid center x"] == 1.05e-05
    assert data1.metadata["grid center y"] == 1.1499999999999998e-05


def test_map_open():
    tf = data_path / "fmt-afm-workshop-fd_mapping_16_2018-08-01_13.07.zip"
    k = 20
    sens = .01e-6
    data = afmformats.load_data(tf, meta_override={"spring constant": k,
                                                   "sensitivity": sens})
    assert len(data) == 16
    # these changes were made manually
    offset = 13761.9288
    for ii in range(len(data)):
        assert data[ii]["height (measured)"][0] == -(offset + ii) * 1e-9

    assert data[0]["height (measured)"][0] != data[1]["height (measured)"][0]
    assert np.allclose(data[0]["force"][0], 0.6875 * sens * k)


def test_missing_sens():
    tf = data_path / "fmt-afm-workshop-fd_single_2018-08-01_13.06.09.csv"
    try:
        afmformats.load_data(tf)
    except afmformats.errors.MissingMetaDataError:
        pass


def test_single_conversion():
    tf = data_path / "fmt-afm-workshop-fd_single_2018-08-01_13.06.09.csv"
    k = 20
    sens = .01e-6
    data = afmformats.load_data(tf, meta_override={"spring constant": k,
                                                   "sensitivity": sens})[0]
    assert data.metadata["spring constant"] == k
    assert "force" in data.columns
    assert np.allclose(data["force"][0], 0.6861 * sens * k)


def test_single_open():
    tf = data_path / "fmt-afm-workshop-fd_single_2020-02-14_13.41.25.csv"
    data = afmformats.load_data(tf)[0]
    assert "force" in data.columns
    assert np.allclose(data["force"][0], 1276.4373e-9)
    assert data.metadata["date"] == "2020-02-14"
    assert data.metadata["time"] == "13:41:26"


def test_single_open_issue_17():
    tf = data_path / "fmt-afm-workshop-fd_single_2021-01-15.csv"
    k = 20
    sens = .01e-6
    data = afmformats.load_data(tf, meta_override={"spring constant": k,
                                                   "sensitivity": sens})[0]
    assert "force" in data.columns
    assert np.allclose(data["force"][0], 1.6524e-07)
    assert data.metadata["spring constant"] == k
    assert data.metadata["software version"] == "4.0.0.62"
    assert data.metadata["date"] == "2021-01-15"
    assert data.metadata["time"] == "13:11:45"


def test_single_open_issue_17_no_or_invalid_metadata():
    tf = data_path / "fmt-afm-workshop-fd_single_2021-01-15.csv"
    with pytest.raises(afmformats.errors.MissingMetaDataError,
                       match="specify sensitivity and spring constant"):
        afmformats.load_data(tf)[0]


def test_single_open_issue_17_valid_metadata():
    tf = data_path / "fmt-afm-workshop-fd_single_2021-01-15.csv"
    tmpd = pathlib.Path(tempfile.mkdtemp("modified"))
    tf2 = tmpd / tf.name

    rawdata = tf.read_text()
    rawdata = rawdata.replace("Light Lever Gain, mV/nm:	1.000000",
                              "Light Lever Gain, mV/nm:	100")
    rawdata = rawdata.replace("Force Constant, nN/nm:	1.000000",
                              "Force Constant, nN/nm:	20.0")

    tf2.write_text(rawdata)

    data = afmformats.load_data(tf2)[0]

    assert "force" in data.columns
    assert data.metadata["spring constant"] == 20
    assert data.metadata["sensitivity"] == .01e-6


def test_single_open_with_precomputed_force():
    # This file contains force data in Newtons but no tip position.
    # afmformats should not request additional metadata, although
    # nanite may, because it needs it for the computation of the
    # tip position.
    tf = data_path / "fmt-afm-workshop-fd_single_2021-10-22_14.16.csv"
    data = afmformats.load_data(tf)[0]
    assert "tip position" not in data
    assert "force" in data
    assert "spring constant" not in data.metadata


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""Test of basic opening functionalities"""
import pathlib

import numpy as np
import pytest


import afmformats.errors
from afmformats.formats.fmt_jpk import jpk_data, load_jpk
from afmformats.formats.fmt_jpk.jpk_reader import ArchiveCache, JPKReader


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_open_jpk_simple():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    jpkr = JPKReader(jpkfile)
    p_seg = jpkr.get_index_segment_path(0, 0)
    loc_list = [ff for ff in jpkr.files if ff.count(p_seg)]
    name, slot, dat = jpk_data.find_column_dat(loc_list, "height (piezo)")
    properties = jpkr._get_index_segment_properties(0, 0)
    arc = ArchiveCache.get(jpkr.path)
    with arc.open(dat, "r") as fd:
        height = jpk_data.load_dat_raw(fd, name=name,
                                       properties=properties)
    assert height[0] == 50.425720226584403


def test_open_jpk_calibration():
    cf = data_path / \
         "fmt-jpk-cl_calibration_force-save-2015.02.04-11.25.21.294.jpk-force"

    with pytest.raises(afmformats.errors.MissingMetaDataError,
                       match="sensitivity"):
        load_jpk(cf)


def test_open_jpk_conversion():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    jpkr = JPKReader(jpkfile)
    p_seg = jpkr.get_index_segment_path(0, 0)
    loc_list = [ff for ff in jpkr.files if ff.count(p_seg)]
    properties = jpkr._get_index_segment_properties(0, 0)
    chan_data = {}
    for column in ["force", "height (piezo)", "height (measured)"]:
        name, slot, dat = jpk_data.find_column_dat(loc_list, column)
        arc = ArchiveCache.get(jpkr.path)
        with arc.open(dat, "r") as fd:
            chan_data[name] = jpk_data.load_dat_unit(fd, name, properties,
                                                     slot)
    assert chan_data["vDeflection"][2] == "vDeflection (Force)"
    assert chan_data["vDeflection"][1] == "N"
    assert chan_data["vDeflection"][0][0] == -5.145579192349918e-10
    assert chan_data["height"][0][0] == 2.8783223430683289e-05
    assert chan_data["strainGaugeHeight"][0][0] == 2.2815672438768612e-05


def test_get_both_metadata():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    jpkr = JPKReader(jpkfile)
    md1 = jpkr.get_metadata(index=0, segment=0)
    md2 = jpkr.get_metadata(index=0, segment=1)
    md = jpkr.get_metadata(index=0, segment=None)
    for key1 in md1:
        if key1 in ["duration", "point count"]:
            assert md1[key1] != md[key1], key1
        else:
            assert md1[key1] == md[key1], key1
    for key2 in md2:
        if key2 in ["duration", "point count", "time"]:
            # time from approach curve should be used
            assert md2[key2] != md[key2], key2
        else:
            assert md2[key2] == md[key2], key2


def test_get_single_curves():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    jpkr = JPKReader(jpkfile)
    jpkr.get_data(column="force", index=0, segment=0)
    jpkr.get_data(column="height (piezo)", index=0, segment=0)
    jpkr.get_data(column="height (measured)", index=0, segment=0)


def test_get_time():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    jpkr = JPKReader(jpkfile)
    time1 = jpkr.get_data(column="time", index=0, segment=0)
    time2 = jpkr.get_data(column="time", index=0, segment=1)
    m1 = jpkr.get_metadata(0, 0)
    m2 = jpkr.get_metadata(0, 1)
    assert np.allclose(m1["duration"], time2[0])
    assert len(time1) == m1["point count"]
    assert len(time2) == m2["point count"]
    assert np.all(time1 < m1["duration"])
    assert np.all(time2 < (m1["duration"] + m2["duration"]))
    md = jpkr.get_metadata(0)
    assert md["duration"] == m1["duration"] + m2["duration"]
    assert md["point count"] == m1["point count"] + m2["point count"]


def test_get_single_custom_slot():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    jpkr = JPKReader(jpkfile)
    p_seg = jpkr.get_index_segment_path(0, 0)
    loc_list = [ff for ff in jpkr.files if ff.count(p_seg)]
    name, slot, dat = jpk_data.find_column_dat(loc_list, "height (piezo)")
    properties = jpkr._get_index_segment_properties(0, 0)
    arc = ArchiveCache.get(jpkr.path)
    with arc.open(dat, "r") as fd:
        data, unit, _ = jpk_data.load_dat_unit(fd, name=name,
                                               properties=properties,
                                               slot="nominal")
    assert data[0] == 4.9574279773415606e-05


def test_meta():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    jpkr = JPKReader(jpkfile)
    md = jpkr.get_metadata(0, 1)
    assert md["spring constant"] == 0.043493666407368466


def test_meta_missing():
    jpkfile = data_path / "fmt-jpk-fd_single-modified_2023.jpk-force"

    with pytest.raises(afmformats.errors.MissingMetaDataError,
                       match="spring constant"):
        load_jpk(jpkfile)

    data_list = load_jpk(jpkfile,
                         meta_override={"spring constant": 12})
    meta = data_list[0]["metadata"]
    assert meta["spring constant"] == 12
    # The sensitivity is stored in the file already.
    assert np.allclose(data_list[0]["data"]["force"][0],
                       -5.261711536426128e-07,
                       atol=1e-12)


def test_meta_override_multiple_times():
    jpkfile = data_path / "fmt-jpk-fd_single-modified_2023.jpk-force"
    jpkr = JPKReader(jpkfile)

    with pytest.raises(afmformats.errors.MissingMetaDataError,
                       match="spring constant"):
        jpkr.get_metadata(index=0)

    jpkr.set_metadata({"spring constant": 10})
    assert jpkr.get_metadata(index=0)["spring constant"] == 10

    jpkr.set_metadata({"spring constant": 12})
    assert jpkr.get_metadata(index=0)["spring constant"] == 12


def test_load_jpk():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    jpkr = JPKReader(jpkfile)
    md = jpkr.get_metadata(0, 1)
    assert md["imaging mode"] == "force-distance"
    assert len(jpkr) == 1, "Only one measurement"
    assert len(jpkr.get_index_segment_numbers(0)) == 2, "approach and retract"


def test_load_jpk_with_own_metadata():
    jpkfile = data_path / "fmt-jpk-fd_single-modified_2023.jpk-force"
    data = afmformats.load_data(jpkfile,
                                meta_override={"spring constant": 10,
                                               "sensitivity": .05})
    apret = data[0]
    assert np.allclose(apret["force"][0], -0.2454449015321114)



import pathlib
import tempfile

import h5py
import numpy as np

import afmformats


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_data_content():
    jpkfile = data_path / "fmt-jpk-sr_cell1-0008.jpk-force"
    fdat = afmformats.load_data(jpkfile)[0]

    assert np.allclose(np.mean(fdat.prep["force"]), -1.1667527118459774e-07,
                       atol=0, rtol=1e-12)
    assert np.allclose(np.mean(fdat.appr["force"]), 1.2185861117864535e-08,
                       atol=0, rtol=1e-12)
    assert np.allclose(np.mean(fdat.intr["force"]), 2.3613889957445826e-08,
                       atol=0, rtol=1e-12)
    assert np.allclose(np.mean(fdat.retr["force"]), -1.4864363676098427e-07,
                       atol=0, rtol=1e-12)

    assert np.allclose(np.mean(fdat.prep["time"]), 0.498046875,
                       atol=0, rtol=1e-12)
    assert np.allclose(np.mean(fdat.appr["time"]), 1.499755859375,
                       atol=0, rtol=1e-12)
    assert np.allclose(np.mean(fdat.intr["time"]), 16.998046875,
                       atol=0, rtol=1e-12)
    assert np.allclose(np.mean(fdat.retr["time"]), 32.47607421875,
                       atol=0, rtol=1e-12)

    assert np.allclose(np.mean(fdat.retr["time"][20]), 32.009765625,
                       atol=0, rtol=1e-12)
    assert np.allclose(np.mean(fdat.appr["force"][1]), -1.1663463779185065e-07,
                       atol=0, rtol=1e-12)


def test_export_hdf5():
    jpkfile = data_path / "fmt-jpk-sr_cell1-0008.jpk-force"
    fdat = afmformats.load_data(jpkfile, modality="stress-relaxation")[0]
    _, path = tempfile.mkstemp(suffix=".h5", prefix="afmformats_test_")

    with h5py.File(path, "w") as h5:
        fdat.export_data(h5, metadata=True, fmt="hdf5")
        fdat.export_data(h5, metadata=True, fmt="hdf5")

    with h5py.File(path, "r") as h5:
        assert h5["0"].attrs["enum"] == 0
        assert h5["1"].attrs["enum"] == 1
        assert h5["0"]["force"].attrs["unit"] == "N"
        assert h5["0"].attrs["imaging mode"] == "stress-relaxation"


def test_load_custom_class():
    class TESTAFMData(afmformats.AFMStressRelaxation):
        pass
    jpkfile = data_path / "fmt-jpk-sr_cell1-0008.jpk-force"
    fdat = afmformats.load_data(jpkfile,
                                modality="stress-relaxation",
                                data_classes_by_modality={
                                    "stress-relaxation": TESTAFMData
                                }
                                )[0]
    assert isinstance(fdat, TESTAFMData)


def test_segment():
    jpkfile = data_path / "fmt-jpk-sr_cell1-0008.jpk-force"
    fdat = afmformats.load_data(jpkfile, modality="stress-relaxation")[0]

    s0 = fdat.prep["segment"]
    assert np.all(s0 == 0)

    s1 = fdat.appr["segment"]
    assert np.all(s1 == 1)

    s2 = fdat.intr["segment"]
    assert np.all(s2 == 2)

    s2 = fdat.retr["segment"]
    assert np.all(s2 == 3)



import numpy as np

from ...errors import FileFormatNotSupportedError


__all__ = ["JPK_COLUMNS",
           "JPK_SLOTS",
           "JPK_UNITS",
           "ReadJPKError",
           "find_column_dat",
           "load_dat_raw",
           "load_dat_unit",
           ]

#: Maps afmformats column names to JPK column names
JPK_COLUMNS = {
    "force": ["vDeflection"],
    "height (measured)": ["strainGaugeHeight", "capacitiveSensorHeight",
                          "measuredHeight"],
    "height (piezo)": ["height", "head-height"],
}

#: Maps afmformats column names to default JPK normalization slots
JPK_SLOTS = {
    "force": "force",
    "height (measured)": "nominal",
    "height (piezo)": "calibrated",
}

#: Maps afmformats column names to default JPK units
JPK_UNITS = {
    "force": "N",
    "height (measured)": "m",
    "height (piezo)": "m",
}


class ReadJPKError(FileFormatNotSupportedError):
    pass


def find_column_dat(loc_list, column):
    """Find a column in a list of strings

    Parameters
    ----------
    loc_list: list of str
        The segment's data location list (within the archive), e.g.

        ['segments/0/channels/height.dat',
         'segments/0/channels/vDeflection.dat',
         'segments/0/channels/strainGaugeHeight.dat']

    column: str
        afmformats column name :const:`afmformats.afm_data.known_columns`

    Returns
    -------
    name: str
        Matched column name from :const:`JPK_COLUMNS`,
        e.g. "strainGaugeHeight"
    slot: str
        Default slot location from :const:`JPK_SLOTS`
    loc: str
        Matched data location in zip file,
        e.g. "segments/0/channels/strainGaugeHeight.dat"
    """
    id_list = JPK_COLUMNS[column]
    col = None
    for dd in loc_list:
        cn = dd.rsplit("/")[-1].rsplit(".")[0]
        for vc in id_list:
            if vc == cn:
                col = vc
                break
        if col:
            break
    else:
        msg = "No data found for: {}".format(id_list)
        raise ReadJPKError(msg)
    slot = JPK_SLOTS[column]
    return col, slot, dd


def get_property(description, keys, properties):
    """Helper function to access data from property dictionaries

    Parameters
    ----------
    description: str
        The human readable name of the property key. Used for
        error handling only to give the user an idea of what went
        wrong.
    keys: list of str
        List of potential property keys
    properties: dict
        Property dictionary metadata (see also
        :func:`JPKReader._get_index_segment_properties`)

    Raises
    ------
    ReadJPKError if a key could not be found in `property_dict`
    """
    for mk in keys:
        if mk in properties:
            mult = properties[mk]
            break
    else:
        raise ReadJPKError(f"Could not find property: '{description}'.\n"
                           "Please make sure you are working with raw data "
                           "and not with data exported from the JPK software.")
    return mult


def load_dat_raw(fd, name, properties):
    """Load data from binary JPK .dat files

    Parameters
    ----------
    fd: file
        Open .dat file
    name: str
        Name of the data to read (required for scale conversions)
        (valid options are values in :const:`JPK_COLUMNS`)
    properties: dict
        Property dictionary metadata (see also
        :func:`JPKReader._get_index_segment_properties`)

    Returns
    -------
    data: 1d ndarray
        A numpy array with the raw data.

    Notes
    -----
    This method tries to correctly determine the data type of the
    binary data and scales it with the `data.encoder.scaling`
    values given in the header files.

    See Also
    --------
    load_dat_unit: Includes conversion to useful units
    """
    # Multiplier
    mult = get_property(
        description=f"{name} multiplier in {fd}",
        keys=[f"channel.{name}.data.encoder.scaling.multiplier",
              f"channel.{name}.encoder.scaling.multiplier"],
        properties=properties)

    # Offset
    off = get_property(
        description=f"{name} offset in {fd}",
        keys=[f"channel.{name}.data.encoder.scaling.offset",
              f"channel.{name}.encoder.scaling.offset"],
        properties=properties)

    # Data type
    enc = get_property(
        description=f"{name} offset in {fd}",
        keys=[f"channel.{name}.data.encoder.type",
              f"channel.{name}.encoder.type"],
        properties=properties)

    # determine encoder
    if enc == "signedshort":
        mydtype = np.dtype(">i2")
    elif enc == "unsignedshort":
        mydtype = np.dtype(">u2")
    elif enc == "signedinteger":
        mydtype = np.dtype(">i4")
    elif enc == "unsignedinteger":
        mydtype = np.dtype(">u4")
    elif enc == "signedlong":
        mydtype = np.dtype(">i8")
    else:
        raise NotImplementedError("Data file format '{}' not supported".
                                  format(enc))

    data = np.frombuffer(fd.read(), dtype=mydtype) * mult + off
    return data


def load_dat_unit(fd, name, properties, slot="default"):
    """Load data from a JPK .dat file with a specific calibration slot

    Parameters
    ----------
    fd: file
        Open .dat file
    name: str
        Name of the data to read (required for scale conversions)
        (valid options are values in :const:`JPK_COLUMNS`)
    properties: dict
        Property dictionary metadata (see also
        :func:`JPKReader._get_index_segment_properties`)
    slot: str
        The .dat files in the JPK measurement zip files come with different
        calibration slots. Valid values are

            - For the height of the piezo crystal during measurement
              (the piezo height is not as accurate as the measured height
              from the height sensor; the piezo movement is not linear):
              "height.dat": "volts", "nominal", "calibrated"

            - For the measured height of the cantilever:
              "strainGaugeHeight.dat": "volts", "nominal", "absolute"
              "measuredHeight.dat":  "volts", "nominal", "absolute"
              "capacitiveSensorHeight": "volts", "nominal", "absolute"
              (they are all the same)

            - For the recorded cantilever deflection:
              "vDeflection.dat": "volts", "distance", "force"

    Returns
    -------
    data: 1d ndarray
        A numpy array containing the scaled data.
    unit: str
        A string representing the metric unit of the data.
    name: str
        The name of the data column.


    Notes
    -----
    The raw data (see `load_dat_raw`) is usually stored in "volts" and
    needs to be converted to e.g. "force" for "vDeflection" or "nominal"
    for "strainGaugeHeight". The conversion parameters (offset, multiplier)
    are stored in the header files and they are not stored separately for
    each slot, but the conversion parameters are stored relative to the
    slots. For instance, to compute the "force" slot from the raw "volts"
    data, one first needs to compute the "distance" slot. This conversion
    is taken care of by this method.

    This is an example header:

        channel.vDeflection.data.file.name=channels/vDeflection.dat
        channel.vDeflection.data.file.format=raw
        channel.vDeflection.data.type=short
        channel.vDeflection.data.encoder.type=signedshort
        channel.vDeflection.data.encoder.scaling.type=linear
        channel.vDeflection.data.encoder.scaling.style=offsetmultiplier
        channel.vDeflection.data.encoder.scaling.offset=-0.00728873489143207
        channel.vDeflection.data.encoder.scaling.multiplier=3.0921021713588157E-4
        channel.vDeflection.data.encoder.scaling.unit.type=metric-unit
        channel.vDeflection.data.encoder.scaling.unit.unit=V
        channel.vDeflection.channel.name=vDeflection
        channel.vDeflection.conversion-set.conversions.list=distance force
        channel.vDeflection.conversion-set.conversions.default=force
        channel.vDeflection.conversion-set.conversions.base=volts
        channel.vDeflection.conversion-set.conversion.volts.name=Volts
        channel.vDeflection.conversion-set.conversion.volts.defined=false
        channel.vDeflection.conversion-set.conversion.distance.name=Distance
        channel.vDeflection.conversion-set.conversion.distance.defined=true
        channel.vDeflection.conversion-set.conversion.distance.type=simple
        channel.vDeflection.conversion-set.conversion.distance.comment=Distance
        channel.vDeflection.conversion-set.conversion.distance.base-calibration-slot=volts
        channel.vDeflection.conversion-set.conversion.distance.calibration-slot=distance
        channel.vDeflection.conversion-set.conversion.distance.scaling.type=linear
        channel.vDeflection.conversion-set.conversion.distance.scaling.style=offsetmultiplier
        channel.vDeflection.conversion-set.conversion.distance.scaling.offset=0.0
        channel.vDeflection.conversion-set.conversion.distance.scaling.multiplier=7.000143623002982E-8
        channel.vDeflection.conversion-set.conversion.distance.scaling.unit.type=metric-unit
        channel.vDeflection.conversion-set.conversion.distance.scaling.unit.unit=m
        channel.vDeflection.conversion-set.conversion.force.name=Force
        channel.vDeflection.conversion-set.conversion.force.defined=true
        channel.vDeflection.conversion-set.conversion.force.type=simple
        channel.vDeflection.conversion-set.conversion.force.comment=Force
        channel.vDeflection.conversion-set.conversion.force.base-calibration-slot=distance
        channel.vDeflection.conversion-set.conversion.force.calibration-slot=force
        channel.vDeflection.conversion-set.conversion.force.scaling.type=linear
        channel.vDeflection.conversion-set.conversion.force.scaling.style=offsetmultiplier
        channel.vDeflection.conversion-set.conversion.force.scaling.offset=0.0
        channel.vDeflection.conversion-set.conversion.force.scaling.multiplier=0.043493666407368466
        channel.vDeflection.conversion-set.conversion.force.scaling.unit.type=metric-unit
        channel.vDeflection.conversion-set.conversion.force.scaling.unit.unit=N

    To convert from the raw "volts" data to force data, these steps are
    performed:

    - Convert from "volts" to "distance" first, because the
      "base-calibration-slot" for force is "distance".

      distance = volts*7.000143623002982E-8 + 0.0

    - Convert from "distance" to "force":

      force = distance*0.043493666407368466 + 0.0

    The multipliers shown above are the values for sensitivity and spring
    constant:
    sensitivity = 7.000143623002982E-8 m/V
    spring_constant = 0.043493666407368466 N/m
    """
    data = load_dat_raw(fd, name=name, properties=properties)

    conv = f"channel.{name}.conversion-set"
    if slot == "default":
        slot = properties[f"{conv}.conversions.default"]

    # get base unit
    base = properties[f"{conv}.conversions.base"]

    # Now iterate through the conversion sets until we have the base converter.
    # A list of multipliers and offsets
    converters = []
    curslot = slot

    while curslot != base:
        # Get current slot multipliers and offsets
        off = properties[f"{conv}.conversion.{curslot}.scaling.offset"]
        mult = properties[f"{conv}.conversion.{curslot}.scaling.multiplier"]
        converters.append([mult, off])
        curslot = properties[
            f"{conv}.conversion.{curslot}.base-calibration-slot"
        ]

    # Get raw data
    for c in converters[::-1]:
        data[:] = c[0] * data[:] + c[1]

    if base == slot:
        unit = properties[f"channel.{name}.data.encoder.scaling.unit.unit"]
    else:
        unit = get_property(
            description=f"scale conversion {name} for {slot} in {fd}",
            keys=[f"{conv}.conversion.{slot}.scaling.unit",
                  f"{conv}.conversion.{slot}.scaling.unit.unit"],
            properties=properties)

    out_name = properties[f"{conv}.conversion.{slot}.name"]
    return data, unit, f"{name} ({out_name})"



from .ws_single import load_csv
from .ws_map import load_map


recipe_workshop_map = {
    "descr": "QMAP as zipped comma-separated values",
    "loader": load_map,
    "suffix": ".zip",
    "modalities": ["force-distance"],
    "maker": "AFM workshop",
}

recipe_workshop_single = {
    "descr": "comma-separated values",
    "loader": load_csv,
    "suffix": ".csv",
    "modalities": ["force-distance"],
    "maker": "AFM workshop",
}



import json
import pathlib

import numpy as np

from ..afm_data import column_dtypes, known_columns


__all__ = ["load_tab"]


def detect_tab(path, max_header=1000):
    """Test whether `path` is in the tab format"""
    has_begin = False
    has_data = False
    has_end = False
    with path.open() as fd:
        for _ in range(max_header):
            line = fd.readline().strip()
            if line.startswith("# BEGIN METADATA"):
                has_begin = True
            elif line.startswith("# END METADATA") and has_begin:
                has_end = True
            elif len(line) == 0 or line.startswith("#"):
                continue
            else:
                # make sure the first line contains actual floats
                if line.count("\t"):
                    has_data = True
                break
    return has_begin and has_data and has_end


def load_tab(path, callback=None, meta_override=None):
    """Loads tab-separated-value files as exported by afmformats

    This is a simple tab-separated values files. The metadata
    may be present at the beginning of the file, commented out,
    as a json dump in a "BEGIN METADATA" - "END METADATA" block.
    The column data is listed below as a simple table.

    Parameters
    ----------
    path: str or pathlib.Path or io.TextIOBase
        path to a .tab file
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)
    """
    if meta_override is None:
        meta_override = {}
    else:
        # just make sure nobody expects a different result for the forces
        for key in ["sensitivity", "spring constant"]:
            if key in meta_override:
                raise NotImplementedError(
                    f"Setting metadata such as '{key}' is not implemented!")

    path = pathlib.Path(path)
    with path.open() as fd:
        tsvdata = fd.readlines()

    # get the metadata
    dump = []
    injson = False
    for ii, line in enumerate(tsvdata):
        if line.startswith("# BEGIN METADATA"):
            injson = True
            continue
        elif line.startswith("# END METADATA"):
            break
        elif injson:
            dump.append(line.strip("#").strip())
    if dump:
        metadata = json.loads("\n".join(dump))
    else:
        metadata = {}
    metadata["path"] = path
    metadata["enum"] = 0

    # last line with a hash is the header
    for ii, line in enumerate(tsvdata):
        if not line.strip():
            # empty line
            pass
        elif line.startswith("#"):
            # header candidate
            header_line = line
        else:
            if ii == 0:
                raise ValueError("No header found in '{}'!".format(path))
            break
    else:
        raise ValueError("No data found in '{}'!".format(path))
    columns = header_line.strip("#").strip().split("\t")

    # load the data
    da = [f.strip() for f in tsvdata if f.strip() and not f.startswith("#")]
    # generate arrays
    data = {}
    for cc in columns:
        if cc in known_columns:
            data[cc] = np.zeros(len(da), dtype=column_dtypes[cc])
    for ii, line in enumerate(da):
        for jj, item in enumerate(line.strip().split("\t")):
            assert jj < len(columns)
            cc = columns[jj]
            if cc in known_columns:
                data[cc][ii] = string_to_dtype(item, column_dtypes[cc])

    metadata.update(meta_override)
    dd = {"data": data,
          "metadata": metadata}

    if callback is not None:
        callback(1)

    return [dd]


def string_to_dtype(astring, dtype):
    astring = astring.strip()
    if astring == "False":
        astring = "0"
    elif astring == "True":
        astring = "1"
    if dtype in [float, int, np.uint8]:
        return dtype(astring)
    else:
        raise ValueError("No conversion rule for dtype '{}'!".format(dtype))


recipe_tab = {
    "descr": "tab-separated values",
    "detect": detect_tab,
    "loader": load_tab,
    "suffix": ".tab",
    "modalities": ["force-distance"],
    "maker": "afmformats",
}



import io
import pathlib
import warnings

import numpy as np

from ... import errors


class AFMWorkshopFormatWarning(UserWarning):
    pass


__all__ = ["load_csv"]


months = {
    "January": 1,
    "February": 2,
    "March": 3,
    "April": 4,
    "May": 5,
    "June": 6,
    "July": 7,
    "August": 8,
    "September": 9,
    "October": 10,
    "November": 11,
    "December": 12
}


def load_csv(path, callback=None, meta_override=None, mode="single"):
    """Load csv data from AFM workshop

    The files are structured like this::

        Force-Distance Curve
        File Format:    3

        Date:    Wednesday, August 1, 2018
        Time:    1:07:47 PM
        Mode:    Single
        Point:    1
        X, um:    27.250000
        Y, um:    27.250000

        Extend Z-Sense(nm),Extend T-B(V),Retract Z-Sense(nm),Retract T-B(V)
        13777.9288,0.6875,14167.9288,1.0917
        13778.9288,0.6874,14166.9288,1.0722
        13779.9288,0.6876,14165.9288,1.0693
        13780.9288,0.6877,14164.9288,1.0824
        13781.9288,0.6875,14163.9288,1.0989
        ...

    The data for testing was kindly provided by Peter Eaton
    (afmhelp.com).


    Parameters
    ----------
    path: str or pathlib.Path or io.TextIOBase
        data file or an open file in text (not bytes) mode
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)
    mode: str
        curve mode to expect (either "single" or "mapping"); if an
        unexpected mode is found, AFMWorkshopFormatWarning is issued
    """
    if meta_override is None:
        meta_override = {}
    metadata = {}
    # TODO:
    # - only read really the header instead of the entire file to save time
    if isinstance(path, io.IOBase):
        csvdata = path.readlines()
        path.seek(0)  # required for np.loadtxt below
    else:
        path = pathlib.Path(path)
        csvdata = path.read_text(encoding="utf-8").split("\n")
        metadata["path"] = path

    # get the metadata
    for ii, line in enumerate(csvdata):
        if line.count("Force-Distance Curve"):
            metadata["imaging mode"] = "force-distance"
        elif line.startswith("Software Version"):
            metadata["software version"] = line.split(":")[1].strip()
        elif line.startswith("Date:"):
            cdate = parse_date(line.split(":")[1])
            metadata["date"] = cdate
        elif line.startswith("Time:"):
            metadata["time"] = line.split(":", 1)[1]
        elif line.startswith("Mode:"):
            cur_mode = line.split(":")[1].strip().lower()
            if cur_mode != mode:
                warnings.warn(f"Expected '{mode}' curve; got '{cur_mode}'!",
                              AFMWorkshopFormatWarning)
        elif line.startswith("Point:"):
            metadata["enum"] = int(line.split(":")[1])
        elif line.startswith("X, um:"):
            metadata["position x"] = float(line.split(":")[1]) * 1e-6
        elif line.startswith("Y, um:"):
            metadata["position y"] = float(line.split(":")[1]) * 1e-6
        elif line.startswith("Light Lever Gain, mV/nm:"):
            value = float(line.split(":")[1])
            if value != 1:  # ignore default values
                metadata["sensitivity"] = 1 / (value * 1e-3 / 1e-9)
        elif line.startswith("Force Constant, nN/nm:"):
            value = float(line.split(":")[1])
            if value != 1:  # ignore default values
                metadata["spring constant"] = value
        elif line.count(",") >= 3:
            header_index = ii
            header_line = line
            break
    else:
        raise errors.FileFormatNotSupportedError(
            "Could not parse metadata correctly: {}".format(path))

    metadata.update(meta_override)
    req_metadata = ["sensitivity", "spring constant"]
    mis_metadata = [key for key in req_metadata if key not in metadata]
    if mis_metadata:
        fmult = None
    else:
        fmult = metadata["sensitivity"] * metadata["spring constant"]

    if "imaging mode" not in metadata:
        raise errors.FileFormatNotSupportedError(
            "Unknown file format: {}".format(path))

    # load data
    rawdata = np.loadtxt(path, skiprows=header_index+1, delimiter=",")
    segsize = rawdata.shape[0]
    data = {"height (measured)": np.zeros(2*segsize, dtype=float)*np.nan,
            "force": np.zeros(2*segsize, dtype=float)*np.nan,
            "segment": np.concatenate((np.zeros(segsize, dtype=np.uint8),
                                       np.ones(segsize, dtype=np.uint8)))
            }
    columns = header_line.strip().split(",")
    for jj, cc in enumerate(columns):
        if cc == "Extend Z-Sense(nm)":
            data["height (measured)"][:segsize] = -rawdata[:, jj] * 1e-9
        elif cc == "Retract Z-Sense(nm)":
            data["height (measured)"][segsize:] = -rawdata[:, jj] * 1e-9
        elif cc == "Extend T-B(V)":
            if mis_metadata:
                raise errors.MissingMetaDataError(
                    mis_metadata,
                    f"Please specify {' and '.join(mis_metadata)}!")
            else:
                data["force"][:segsize] = rawdata[:, jj] * fmult
        elif cc == "Extend Force(nN)":
            data["force"][:segsize] = rawdata[:, jj] * 1e-9
        elif cc == "Retract T-B(V)":
            if mis_metadata:
                raise errors.MissingMetaDataError(
                    mis_metadata,
                    f"Please specify {' and '.join(mis_metadata)}!")
            else:
                data["force"][segsize:] = rawdata[:, jj] * fmult
        elif cc == "Retract Force(nN)":
            data["force"][segsize:] = rawdata[:, jj] * 1e-9
        else:
            warnings.warn(f"Unknown column encountered: {cc}",
                          AFMWorkshopFormatWarning)

    # remove non-existent or bad data
    for key in list(data.keys()):
        if np.sum(np.isnan(data[key])):
            warnings.warn(f"Removed incomplete column: {key} in {path}!",
                          AFMWorkshopFormatWarning)
            data.pop(key)

    dd = {"data": data,
          "metadata": metadata}

    if callback is not None:
        callback(1)
    return [dd]


def parse_date(datestr):
    """Heuristic function for parsin AFM workshop dates"""
    datestr = datestr.strip()

    if datestr.count(",") == 2:
        # Friday, February 14, 2020
        _, m, d, y = [a.strip(", ") for a in datestr.split()]
    elif datestr.count(",") == 0:
        # 15 January 2021
        d, m, y = [a.strip("") for a in datestr.split()]
    else:
        raise errors.FileFormatMetaDataError(f"Cannot parse date '{datestr}'!")

    return f"{int(y):04d}-{months[m]:02d}-{int(d):02d}"



import pathlib

from .afm_data import AFMData
from .formats import load_data


__all__ = ["AFMGroup"]


class AFMGroup(object):
    """Container for :class:`afmformats.afm_data.AFMData`"""
    def __init__(self, path=None, meta_override=None, callback=None,
                 modality=None, data_classes_by_modality=None):
        """
        Parameters
        ----------
        path: str or pathlib.Path or None
            If this option is specified, then an AFMGroup is generated
            directly from a datafile.
        meta_override: dict
            Dictionary with metadata that is used when loading the data
            in `path`.
        callback: callable or None
            A method that accepts a float between 0 and 1
            to externally track the process of loading the data.
        data_classes_by_modality: dict
            Override the default AFMData class to use for managing the data
            (see :data:`default_data_classes_by_modality`): This is e.g.
            used by :ref:`nanite:index` to pass `Indentation` (which is a
            subclass of the default `AFMForceDistance`) for handling
            "force-indentation" data.
        """
        if path is not None:
            path = pathlib.Path(path)
        self._mmlist = []

        if path is not None:
            self += load_data(
                path,
                callback=callback,
                meta_override=meta_override,
                modality=modality,
                data_classes_by_modality=data_classes_by_modality
            )
        elif meta_override is not None:
            raise ValueError("Specifying `meta_override` without specifying "
                             "`path` is meaningless.")

        self.path = path

    def __add__(self, grp):
        out = AFMGroup()
        out += self
        out += grp
        return out

    def __iadd__(self, grp):
        if not isinstance(grp, (AFMGroup, list)):
            raise ValueError("Please specify list or AFMGroup for iadd!")
        for afmdata in grp:
            self.append(afmdata)
        self.path = None
        return self

    def __iter__(self):
        return iter(self._mmlist)

    def __getitem__(self, idx):
        return self._mmlist[idx]

    def __len__(self):
        return len(self._mmlist)

    def __repr__(self):
        return f"<{self.__class__.__name__}: '{self.path}' at {hex(id(self))}>"

    def __str__(self):
        rep = [f"{self.__class__.__name__}: '{self.path}'"]
        for afmdata in self._mmlist:
            rep.append("- {}".format(afmdata))
        return "\n".join(rep)

    def append(self, afmdata):
        """Append an AFMData instance

        Parameters
        ----------
        afmdata: afmformats.afm_data.AFMData
            AFM data
        """
        if not isinstance(afmdata, AFMData):
            raise ValueError("`afmdata` must be an instance of `AFMData`!")
        self._mmlist.append(afmdata)

    def get_enum(self, enum):
        """Return the AFMData curve with this enum value

        Raises
        ------
        ValueError if multiple curves with the same enum value exist.
        KeyError if the enum value is not found
        """
        curves = []
        for afmdata in self._mmlist:
            if afmdata.enum == enum:
                curves.append(afmdata)
        if len(curves) == 0:
            raise KeyError(f"Could not find dataset with enum '{enum}'!")
        elif len(curves) == 1:
            return curves[0]
        else:
            raise ValueError(
                f"Multiple curves with the enum value '{enum}' exist!")

    def index(self, afmdata):
        return self._mmlist.index(afmdata)

    def subgroup_with_path(self, path):
        """Return a subgroup with AFMData matching `path`"""
        path = pathlib.Path(path)
        subgroup = AFMGroup()
        for afmdata in self:
            if afmdata.path.samefile(path):
                subgroup.append(afmdata)
        subgroup.path = path
        return subgroup



import abc
import io
import json
import pathlib
import warnings

import h5py
import numpy as np

from ._version import version
from .meta import MetaData


__all__ = ["AFMData", "column_dtypes", "column_units", "known_columns"]


class AFMData(abc.ABC):
    """General base class for AFM data"""

    def __init__(self, data, metadata, diskcache=False):
        """Initialization

        Parameters
        ----------
        data: dict-like
            Experimental data
        metadata: dict
            Metadata
        diskcache: bool
            TODO
        """
        # convert meta data
        metadata_i = MetaData(metadata)
        # check data keys
        for cc in data:
            if cc not in known_columns:
                raise ValueError("Unknown column name '{}'!".format(cc))
        self._path = pathlib.Path(metadata_i["path"])
        self._metadata = metadata_i
        self._enum = metadata_i["enum"]
        # raw data will not be touched
        self._raw_data = data
        self._data = {}

    def __contains__(self, key):
        return self._data.__contains__(key) or self._raw_data.__contains__(key)

    def __getitem__(self, key):
        if key in self._data:
            data = self._data[key]
        elif key in self._raw_data:
            data = self._raw_data[key].copy()
        elif key == "index":
            return np.arange(len(self))
        else:
            raise KeyError("Column '{}' not defined!".format(key))
        return data

    def __len__(self):
        # If you are here, you might have asked yourself why
        # loading your data takes so long. You have tried
        # lazy-loading and there shouldn't be a reason why
        # it is sooo slow! Except there is, because we need
        # the size of the dataset - and if that is not in the
        # metadata, then we just take the length of the first
        # data column as a workaround (which takes time).
        if "point count" not in self._metadata:
            k0 = list(self._raw_data.keys())[0]
            length = len(self[k0])
        else:
            length = self._metadata["point count"]
        return length

    def __setitem__(self, key, values):
        """Set column data"""
        if len(values) != len(self):
            raise ValueError(
                f"Cannot set data '{key}' of length '{len(values)}' "
                + f"for AFMForceDistance of length '{len(self)}'!")
        # do not touch raw data
        self._data[key] = values

    def __str__(self):
        strre = "{} '{}'[{}]".format(
            self.__class__.__name__, self.path, self.enum)
        return strre

    def __repr__(self):
        repre = "<{} '{}'[{}] at {}>".format(
            self.__class__.__name__, self.path, self.enum, hex(id(self)))
        return repre

    @property
    def columns(self):
        """Available data columns"""
        raw = self._raw_data.keys()
        new = self._data.keys()
        return sorted(set(raw) | set(new))

    @property
    def columns_innate(self):
        """Data columns available only in the original data file"""
        return sorted(self._raw_data.keys())

    @property
    def enum(self):
        """Unique index of `self` in `self.path`

        Indexing starts at "0"
        """
        return self._enum

    @property
    def metadata(self):
        """Unique index of `self` in `self.path`"""
        return MetaData(self._metadata.copy())

    @property
    @abc.abstractmethod
    def modality(self):
        """Imaging modality (e.g. force-distance)"""

    @property
    def path(self):
        """Path to the measurement file"""
        return self._path

    def _export_hdf5(self, h5group, metadata_dict=None):
        """Export data to the HDF5 file format

        Parameters
        ----------
        h5group: h5py.Group or h5py.File
            Destination group
        metadata_dict: dict
            Key-value pairs for the metadata that should be exported
            (will be stored in the group attributes)
        """
        # set the software and its version
        if metadata_dict is None:
            metadata_dict = {}
        h5group.attrs["software"] = "afmformats"
        h5group.attrs["software version"] = version
        enum_key = str(self.enum)
        if enum_key in h5group:
            # random fill-mode (get the next free enum key)
            ii = 0
            while True:
                enum_key = str(ii)
                if enum_key not in h5group:
                    break
                ii += 1
        metadata_dict["enum"] = int(enum_key)
        subgroup = h5group.create_group(enum_key)
        for col in self.columns:
            if col == "segment":
                ds = subgroup.create_dataset(name=col,
                                             data=np.asarray(self[col],
                                                             dtype=np.uint8),
                                             compression="gzip",
                                             fletcher32=True)
            elif col == "index":
                # do not store index column
                continue
            else:
                ds = subgroup.create_dataset(name=col,
                                             data=self[col],
                                             compression="gzip",
                                             fletcher32=True)
            ds.attrs["unit"] = column_units[col]
        for kk in metadata_dict:
            if kk == "path":
                subgroup.attrs["path"] = str(metadata_dict["path"])
            else:
                subgroup.attrs[kk] = metadata_dict[kk]

    def _export_tab(self, fd, metadata_dict=None):
        """Export data to a tab separated values file

        Parameters
        ----------
        fd: io.IOBase
            File opened in "w" mode
        metadata_dict: dict
            Key-value pairs for the metadata that should be exported
            (will be stored in the group attributes)
        """
        if metadata_dict is None:
            metadata_dict = {}
        fd.write("# afmformats {}\r\n".format(version))
        fd.write("#\r\n")
        if metadata_dict:
            # write metadata
            dump = json.dumps(metadata_dict, sort_keys=True, indent=2,
                              default=json_path_serializer)
            fd.write("# BEGIN METADATA\r\n")
            for dl in dump.split("\n"):
                fd.write("# " + dl + "\r\n")
            fd.write("# END METADATA\r\n")
            fd.write("#\r\n")
        # get all data (faster than doing it every time for every row)
        data = {}
        for cc in self.columns:
            data[cc] = self[cc]
        # header
        fd.write("# " + "\t".join(self.columns) + "\r\n")
        # rows
        for ii in range(len(self)):
            items = []
            for cc in self.columns:
                items.append("{:.8g}".format(data[cc][ii]))
            fd.write("\t".join(items) + "\r\n")

    def export(self, *args, **kwargs):
        warnings.warn("Pleas use `export_data` for data export!",
                      DeprecationWarning)
        return self.export_data(*args, **kwargs)

    def export_data(self, out, metadata=True, fmt="tab"):
        """Export all data columns to a file

        Parameters
        ----------
        out: str, pathlib.Path, writable io.IOBase, or h5py.Group
            Output path, open file, or h5py object
        metadata: bool or list
            If True, all available metadata are stored. If False,
            no metadata are stored. If a list, only the given
            metadata keys are stored.
        fmt: str
            "tab" for the tab separated values format and "hdf5" / "h5"
            for the HDF5 file format

        Notes
        -----
        - If you wish to append HDF5 data to an existing file, please
          open the file first and call this function with the h5py.File
          object, i.e.

          .. code:: python

              with h5py.File(path, "a") as h5:
                  fdist.export(out=h5, fmt="hdf5")

          Otherwise the file will be overridden.
        - The column "index" is not exported in the HDF5 file
          format
        """
        if isinstance(metadata, (list, tuple)):
            # list of keys
            metadata_dict = {}
            for key in metadata:
                metadata_dict[key] = self.metadata[key]
        elif isinstance(metadata, bool) and metadata:
            # all metadata
            metadata_dict = self.metadata
        else:
            raise ValueError("Metadata must be list, tuple, or bool, got "
                             f"'{metadata}' of type '{type(metadata)}'!")

        if fmt == "tab":
            if isinstance(out, (pathlib.Path, str)):
                fd = pathlib.Path(out).open("w")
                close = True
            elif isinstance(out, io.IOBase):
                fd = out
                close = False
            else:
                raise ValueError("Unexpected object class for 'out': "
                                 + "'{}' for format 'tab'!".format(
                                     out.__class__))
            self._export_tab(fd, metadata_dict=metadata_dict)
            if close:
                fd.close()
        elif fmt in ["hdf5", "h5"]:
            if isinstance(out, (pathlib.Path, str)):
                # overrides always
                h5 = h5py.File(out, "w")
                close = True
            elif isinstance(out, h5py.Group):
                h5 = out
                close = False
            else:
                raise ValueError("Unexpected object class for 'out': "
                                 + "'{}' for format 'hdf5'!".format(
                                     out.__class__))
            self._export_hdf5(h5group=h5, metadata_dict=metadata_dict)
            if close:
                h5.close()
        else:
            raise ValueError("Unexpected string for 'fmt': {}".format(fmt))

    def reset_data(self):
        """Resets all data to the state they were after loading

        Internally, only `self._data` is `clear`ed, which means
        that all calls to `__getitem__` fall-back to `self._raw_data`.
        """
        self._data.clear()


def json_path_serializer(obj):
    """Used to convert pathlib.Path to str in metadata"""
    if isinstance(obj, pathlib.Path):
        return str(obj)
    else:
        raise TypeError(f"TypeError: Object of type {obj.__class__} "
                        + "is not JSON serializable")


#: Data types of all known columns (all other columns are assumed to be float)
column_dtypes = {
    "force": float,
    "height (measured)": float,
    "height (piezo)": float,
    "index": int,
    "segment": np.uint8,
    "time": float,
    "tip position": float,
}

#: Units of all known columns
column_units = {
    "force": "N",
    "height (measured)": "m",
    "height (piezo)": "m",
    "index": "",
    "segment": "",
    "time": "s",
    "tip position": "m",
}

#: Known data columns
known_columns = sorted(column_dtypes.keys())



class AFMFileFormatError(BaseException):
    pass


class DataFileBrokenError(AFMFileFormatError):
    pass


class FileFormatNotSupportedError(AFMFileFormatError):
    pass


class FileFormatMetaDataError(AFMFileFormatError):
    pass


class InvalidFileFormatError(AFMFileFormatError):
    pass


class MissingMetaDataError(BaseException):
    def __init__(self, meta_keys, *args, **kwargs):
        """Special error class for missing metadata

        The missing metadata keys are stored in the
        ``meta_keys`` property.
        """
        #: List of missing metadata keys
        self.meta_keys = meta_keys
        super(MissingMetaDataError, self).__init__(*args, **kwargs)


__all__ = [e for e in dir() if not e.startswith("__")]



from .afm_data import AFMData
from .afm_segment import AFMSegment


__all__ = ["AFMStressRelaxation"]


class AFMStressRelaxation(AFMData):
    """Base class for AFM stress-relaxation data

    A stress-relaxation dataset consists of a pre-approach (with
    constant height), an approach (with a predefined indentation depth),
    an intermediate (with constant height), and a retract curve.
    """
    def __init__(self, *args, **kwargs):
        super(AFMStressRelaxation, self).__init__(*args, **kwargs)
        #: Dictionary-like interface to the pre-measurement segment
        self.prep = AFMSegment(self._raw_data, self._data, segment=0)
        #: Dictionary-like interface to the approach segment
        self.appr = AFMSegment(self._raw_data, self._data, segment=1)
        #: Dictionary-like interface to the intermediate segment
        self.intr = AFMSegment(self._raw_data, self._data, segment=2)
        #: Dictionary-like interface to the retract segment
        self.retr = AFMSegment(self._raw_data, self._data, segment=3)

    def __setitem__(self, key, value):
        super(AFMStressRelaxation, self).__setitem__(key, value)
        if key == "segment":
            # The user changed the segment, which means we have to clear
            # the segment cache.
            self.prep.clear_cache()
            self.appr.clear_cache()
            self.intr.clear_cache()
            self.retr.clear_cache()

    @property
    def modality(self):
        """Imaging modality"""
        return "stress-relaxation"



from os.path import dirname, exists, realpath
from setuptools import setup, find_packages
import sys

author = "Paul Müller"
authors = [author, "Shada Abuhattum"]
description = 'reading common AFM file formats'
name = 'afmformats'
year = "2019"


sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
try:
    from _version import version  # noqa: F821
except BaseException:
    version = "unknown"


setup(
    name=name,
    author=author,
    author_email='dev@craban.de',
    url='https://github.com/AFM-analysis/afmformats',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    license="MIT",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["h5py",
                      "igor2>=0.5.0",  # Asylum Research .ibw file format
                      "jprops",  # JPK file format
                      "numpy>=1.14.0",
                      ],
    python_requires='>=3.6, <4',
    keywords=["atomic force microscopy",
              "mechanical phenotyping",
              "tissue analysis"],
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Visualization',
        'Intended Audience :: Science/Research'
    ],
    platforms=['ALL'],
)
